
__author__ = 'Lukas Leufen'
__date__ = '2020-07-07'

from collections import Iterator, Iterable
import tensorflow.keras as keras
import numpy as np
import math
import os
import shutil
import psutil
import multiprocessing
import logging
import dill
from typing import Tuple, List


class StandardIterator(Iterator):

    _position: int = None

    def __init__(self, collection: list):
        assert isinstance(collection, list)
        self._collection = collection
        self._position = 0

    def __next__(self):
        """Return next element or stop iteration."""
        try:
            value = self._collection[self._position]
            self._position += 1
        except IndexError:
            raise StopIteration()
        return value


class DataCollection(Iterable):

    def __init__(self, collection: list = None, name: str = None):
        if collection is None:
            collection = []
        assert isinstance(collection, list)
        self._collection = collection.copy()
        self._mapping = {}
        self._set_mapping()
        self._name = name

    @property
    def name(self):
        return self._name

    def __len__(self):
        return len(self._collection)

    def __iter__(self) -> Iterator:
        return StandardIterator(self._collection)

    def __getitem__(self, index):
        if isinstance(index, int):
            return self._collection[index]
        else:
            return self._collection[self._mapping[str(index)]]

    def add(self, element):
        self._collection.append(element)
        self._mapping[str(element)] = len(self._collection) - 1

    def _set_mapping(self):
        for i, e in enumerate(self._collection):
            self._mapping[str(e)] = i

    def keys(self):
        return list(self._mapping.keys())


class KerasIterator(keras.utils.Sequence):

    def __init__(self, collection: DataCollection, batch_size: int, batch_path: str, shuffle_batches: bool = False,
                 model=None, upsampling=False, name=None, use_multiprocessing=False, max_number_multiprocessing=1):
        self._collection = collection
        batch_path = os.path.join(batch_path, str(name if name is not None else id(self)))
        self._path = os.path.join(batch_path, "%i.pickle")
        self.batch_size = batch_size
        self.model = model
        self.shuffle = shuffle_batches
        self.upsampling = upsampling
        self.indexes: list = []
        self._cleanup_path(batch_path)
        self._prepare_batches(use_multiprocessing, max_number_multiprocessing)

    def __len__(self) -> int:
        return len(self.indexes)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        """Get batch for given index."""
        return self.__data_generation(self.indexes[index])

    def _get_model_rank(self):
        if self.model is not None:
            try:
                mod_out = self.model.output_shape
            except AttributeError as e:
                # ToDo replace except statemnet with something meaningful. Depending on BNN architecture the attr
                # output_shape might not be defined. We use it here to check the number of tails -> make sure multiple
                # tails would also work with BNNs in future versions
                mod_out = (None, None)
            if isinstance(mod_out, tuple):  # only one output branch: (None, ahead)
                mod_rank = 1
            elif isinstance(mod_out, list):  # multiple output branches, e.g.: [(None, ahead), (None, ahead)]
                mod_rank = len(mod_out)
            else:  # pragma: no cover
                raise TypeError("model output shape must either be tuple or list.")
            return mod_rank
        else:  # no model provided, assume to use single output
            return 1

    def __data_generation(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        """Load pickle data from disk."""
        file = self._path % index
        with open(file, "rb") as f:
            data = dill.load(f)
        return data["X"], data["Y"]

    @staticmethod
    def _concatenate(new: List[np.ndarray], old: List[np.ndarray]) -> List[np.ndarray]:
        """Concatenate two lists of data along axis=0."""
        return list(map(lambda n1, n2: np.concatenate((n1, n2), axis=0), old, new))

    @staticmethod
    def _concatenate_multi(*args: List[np.ndarray]) -> List[np.ndarray]:
        """Concatenate two lists of data along axis=0."""
        return list(map(lambda *_args: np.concatenate(_args, axis=0), *args))

    def _prepare_batches(self, use_multiprocessing=False, max_process=1) -> None:
        """
        Prepare all batches as locally stored files.

        Walk through all elements of collection and split (or merge) data according to the batch size. Too long data
        sets are divided into multiple batches. Not fully filled batches are retained together with remains from the
        next collection elements. These retained data are concatenated and also split into batches. If data are still
        remaining afterwards, they are saved as final smaller batch. All batches are enumerated by a running index
        starting at 0. A list with all batch numbers is stored in class's parameter indexes. This method can either
        use a serial approach or use multiprocessing to decrease computational time.
        """
        index = 0
        remaining = []
        mod_rank = self._get_model_rank()
        n_process = min([psutil.cpu_count(logical=False), len(self._collection), max_process])  # use only physical cpus
        if n_process > 1 and use_multiprocessing is True:  # parallel solution
            pool = multiprocessing.Pool(n_process)
            output = []
        else:
            pool = None
            output = None
        for data in self._collection:
            length = data.__len__(self.upsampling)
            batches = _get_number_of_mini_batches(length, self.batch_size)
            if pool is None:
                res = f_proc(data, self.upsampling, mod_rank, self.batch_size, self._path, index)
                if res is not None:
                    remaining.append(res)
            else:
                output.append(pool.apply_async(f_proc, args=(data, self.upsampling, mod_rank, self.batch_size, self._path, index)))
            index += batches
        if output is not None:
            for p in output:
                res = p.get()
                if res is not None:
                    remaining.append(res)
            pool.close()
        if len(remaining) > 0:
            X = self._concatenate_multi(*[e[0] for e in remaining])
            Y = self._concatenate_multi(*[e[1] for e in remaining])
            length = X[0].shape[0]
            batches = _get_number_of_mini_batches(length, self.batch_size)
            remaining = f_proc((X, Y), self.upsampling, mod_rank, self.batch_size, self._path, index)
            index += batches
            if remaining is not None:
                _save_to_pickle(self._path, X=remaining[0], Y=remaining[1], index=index)
                index += 1
        self.indexes = np.arange(0, index).tolist()
        if pool is not None:
            pool.join()

    @staticmethod
    def _cleanup_path(path: str, create_new: bool = True) -> None:
        """First remove existing path, second create empty path if enabled."""
        if os.path.exists(path):
            shutil.rmtree(path)
        if create_new is True:
            os.makedirs(path)

    def on_epoch_end(self) -> None:
        """Randomly shuffle indexes if enabled."""
        if self.shuffle is True:
            np.random.shuffle(self.indexes)


def _save_to_pickle(path, X: List[np.ndarray], Y: List[np.ndarray], index: int) -> None:
    """Save data as pickle file with variables X and Y and given index as <index>.pickle ."""
    data = {"X": X, "Y": Y}
    file = path % index
    with open(file, "wb") as f:
        dill.dump(data, f)


def _get_batch(data_list: List[np.ndarray], b: int, batch_size: int) -> List[np.ndarray]:
    """Get batch according to batch size from data list."""
    return list(map(lambda data: data[b * batch_size:(b + 1) * batch_size, ...], data_list))


def _permute_data(X, Y):
    p = np.random.permutation(len(X[0]))  # equiv to .shape[0]
    X = list(map(lambda x: x[p], X))
    Y = list(map(lambda x: x[p], Y))
    return X, Y


def _get_number_of_mini_batches(number_of_samples: int, batch_size: int) -> int:
    """Return number of mini batches as the floored ration of number of samples to batch size."""
    return math.floor(number_of_samples / batch_size)


def f_proc(data, upsampling, mod_rank, batch_size, _path, index):
    if isinstance(data, tuple) is True:
        X, _Y = data
    else:
        X, _Y = data.get_data(upsampling=upsampling)
    Y = [_Y[0] for _ in range(mod_rank)]
    if upsampling:
        X, Y = _permute_data(X, Y)
    length = X[0].shape[0]
    batches = _get_number_of_mini_batches(length, batch_size)
    for b in range(batches):
        batch_X, batch_Y = _get_batch(X, b, batch_size), _get_batch(Y, b, batch_size)
        _save_to_pickle(_path, X=batch_X, Y=batch_Y, index=index)
        index += 1
    if (batches * batch_size) < length:  # keep remaining to concatenate with next data element
        remaining = (_get_batch(X, batches, batch_size), _get_batch(Y, batches, batch_size))
    else:
        remaining = None
    return remaining
